import os
os.environ["OMP_NUM_THREADS"] = "1"
import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import numpy as np
import pandas as pd
import numpy.core.multiarray
import fastcluster
from multiprocessing import Pool,cpu_count
from functools import partial
from contextlib import closing
from joblib import Parallel, delayed
import itertools
import scipy.stats as st

def flatten(container):
    for i in container:
        if isinstance(i, (list,tuple)):
            for j in flatten(i):
                yield j
        else:
            yield i
def flatx(x):
    return tuple(sorted(flatten(x)))

def dist(R,method):
	N = R.shape[0]
	d = R[np.triu_indices(N,1)]


	if method=='average':
		out = fastcluster.average(d)
	if method=='complete':
		out = fastcluster.complete(d)
	if method=='single':
		out = fastcluster.single(d)

	outI = out.astype(int)

	dend = {i:(i,) for i in xrange(N)}

	for i in xrange(len(outI)):
		dend[i+N] = (dend[outI[i][0]],dend[outI[i][1]])


	for i in xrange(N):
		dend.pop(i,None)

	dend = [(flatx(a),flatx(b)) for a,b in dend.values()]

	dend ={flatx((a,b)):(np.array(a),np.array(b)) for a,b in dend}
	return dend
   
def singPV(Rb,method):
    
    if method=='average':
        rxy = dict(map(lambda (k,(a,b)): (k,Rb[a][:,b].mean()),LV.items()))
    if method=='complete':
        rxy = dict(map(lambda (k,(a,b)): (k,Rb[a][:,b].max()),LV.items()))
    if method=='single':
        rxy = dict(map(lambda (k,(a,b)): (k,Rb[a][:,b].min()),LV.items()))

    PV = map(lambda (k,q):int(rxy[q]<rxy[k]), gen.items())
    
    return PV

#X,LV,nan,method,gen
def SingleBoot(nan,method,metric,sel):
    
    Xb = Xg[:,sel]
    if nan==False:
		if metric=='pearson':
			Rb = 1.-np.corrcoef(Xb)
		elif metric=='spearman':
			Rb = 1.-st.spearmanr(Xb.T)[0]
    else:
        Rb = 1.-np.array(pd.DataFrame(Xb.T).corr())
        
    return singPV(Rb,method)

    
def BootDist(method,Nt=1000,nan=False,ncpu=1,metric='pearson'):
	
	global gen
	gen = {tuple(s):k for k in LV for s in LV[k] if len(s)>1}

	sels = np.random.choice(range(Xg.shape[1]),replace=True,size=(Nt,Xg.shape[1]))
	
	f = partial(SingleBoot,nan,method,metric)

	PV = Parallel(n_jobs=ncpu, backend="threading")(delayed(f)(sel) for sel in sels)

	return sorted(zip(np.array(PV).mean(axis=0),gen.keys()))
    
def FDR(PV,alpha,N):
	p = np.array(zip(*PV)[0])
	thr = np.arange(1,len(PV)+1)*alpha/len(PV)

	sel = np.where(p<thr)[0]
	if len(sel)==0: 
		thr=-1.
	else:
		thr = thr[sel][-1]

	L = [range(N)]+[PV[i][1] for i in np.where(p<=thr)[0]]
	L = map(tuple,sorted(L,key=len,reverse=True))
	
	PV = {c:p for p,c in PV}
	PV[tuple(range(N))] = np.nan
	return L,PV

def Analytic_PV(r,n):

	R = 1-r
	gen = {tuple(s):k for k in LV for s in LV[k] if len(s)>1}

	sd = np.vectorize(lambda x: (1./n)*(1-x**2)**2)
	COV = lambda (j,k),(h,m): (0.5/n)*(((R[j,h]-R[j,k]*R[h,k])*(R[k,m]-R[k,h]*R[h,m]))+((R[j,m]-R[j,h]*R[h,m])*(R[k,h]-R[k,j]*R[j,h]))+((R[j,h]-R[j,m]*R[m,h])*(R[k,m]-R[k,j]*R[j,m]))+((R[j,m]-R[j,k]*R[k,m])*(R[k,h]-R[k,m]*R[m,h])))

	PV = []
	for h,(son,fat) in enumerate(gen.items()):
		
		
		a_s,b_s = LV[son]
		a_f,b_f = LV[fat]

		m = R[a_s][:,b_s].mean()-R[a_f][:,b_f].mean()

		ss = sd(R[a_s][:,b_s])
		sf = sd(R[a_f][:,b_f])

		elm_f = list(itertools.product(a_f, b_f))
		elm_s = list(itertools.product(a_s, b_s))

		ns = 1./np.size(ss)
		nf = 1./np.size(sf)

		cf = (nf**2)*2*sum(COV(elm_f[i],elm_f[j]) for i in xrange(len(elm_f)) for j in xrange(i))
		cs = (ns**2)*2*sum(COV(elm_s[i],elm_s[j]) for i in xrange(len(elm_s)) for j in xrange(i))
		cmix = -2*(ns*nf)*sum(COV(elm_f[i],elm_s[j]) for i in xrange(len(elm_f)) for j in xrange(len(elm_s)))
		
		x = (ns**2)*ss.sum()+(nf**2)*sf.sum()+cf+cs+cmix
		
		s = np.sqrt(x)

		PV.append((st.norm.cdf(0,m,s),son))

	return PV
	
def Find_ValidatedCluster(X,Nt=1000,alpha=0.05,nan=False,ncpu=cpu_count(),dendrogram=None,method='average',metric='pearson'):
	if nan==False:
		if metric=='pearson':
			R = 1 - np.corrcoef(X)
		elif metric=='spearman':
			R = 1-st.spearmanr(X.T)[0]
	else:
		R = 1. -np.array(pd.DataFrame(X.T).corr()) 
		
	global Xg
	Xg = X
	
	global LV
	if dendrogram==None:
		LV = dist(R,method)
	else:
		LV = dendrogram
	
	
	if Nt==0:
		PV = Analytic_PV(R,X.shape[1])
	else:	
		PV = BootDist(method,Nt,nan,ncpu,metric)
	
	
	L,PV = FDR(PV,alpha,X.shape[0])

	return L,PV,LV
	
	




import os
os.environ["OMP_NUM_THREADS"] = "1"
import matplotlib.pyplot as plt
import matplotlib.cm as cmap
from metric import StandardDendrogram,OrderArray,ToMemb
import matplotlib.patches as patches
import numpy as np
import seaborn as sns
import pandas as pd
import scipy.stats as st

def DendroAndCorrDist(x,L,LV,method='average',xylabel='Objects',file_w=None,color_style='Set1',nan=False,metric='pearson'):
    
   
	ordn = OrderArray(StandardDendrogram(ToMemb(LV.keys(),x.shape[0])))
	x = x[ordn]

	rev_ornd = dict(sorted(zip(ordn,range(len(ordn)))))

	LV = {tuple(sorted([rev_ornd[k] for k in K])):(np.array(sorted([rev_ornd[a] for a in LV[K][0]])),
												   np.array(sorted([rev_ornd[a] for a in LV[K][1]]))) for K in LV}

	L = [tuple(sorted([rev_ornd[k] for k in K])) for K in L]

	if nan==False:
		if metric=='pearson':
			R = 1 - np.corrcoef(x)
		elif metric=='spearman':
			R = 1 - st.spearmanr(x.T)[0]
	else:
		R = 1. -np.array(pd.DataFrame(x.T).corr()) 
	#R = 1.-np.corrcoef(x)

	L = sorted(L,key=len)

	N = R.shape[0]
	K = LV.keys()

	import matplotlib.patches as patches
	import colorsys
	nc  = len(L)
	import matplotlib as mpl
	import matplotlib.cm as cm

	cnmx = sns.color_palette(color_style, nc-1)

	from collections import defaultdict
	col = defaultdict(int)
	for k in K:
		for j in range(len(L)):
			if set(k) <= set(L[j]):

				if tuple(L[j])==tuple(range(N)):
					col[tuple(k)] = 'black'
					break

				col[tuple(k)] = cnmx[j-1]
				break


	from matplotlib import gridspec

	fig = plt.figure(figsize=(5,9))
	gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1.34]) 

	# the fisrt subplot
	axarr = [0,0]
	axarr[0] = plt.subplot(gs[0])
	axarr[1] = plt.subplot(gs[1], sharex = axarr[0])


	pos = defaultdict(tuple)

	for i in range(N):
		pos[(i,)] = (float(i),0)

	K = sorted(LV.keys(),key=len)

	hm = []
	for k in K:
		d = tuple(sorted(LV[k][0])),tuple(sorted(LV[k][1]))

		c = tuple(sorted(k))
		#if not col.has_key(c):
		#    break
		pp = pos[d[0]][0],pos[d[1]][0]
		h0 = pos[d[0]][1],pos[d[1]][1]

		if method=='average':
			h = R[LV[k][0]][:,LV[k][1]].mean()
		elif method=='complete':
			h = R[LV[k][0]][:,LV[k][1]].max()
		elif method=='single':
			h = R[LV[k][0]][:,LV[k][1]].min()

		axarr[0].plot([pp[0],pp[0]],[h0[0],h],'-',color='k',alpha=0.5)
		axarr[0].plot([pp[1],pp[1]],[h0[1],h],'-',color='k',alpha=0.5)
		axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.5)
		axarr[0].plot([pp[0],pp[1]],[h,h],'-',color=col[c],lw=1.4,zorder=1000)

		hm.append(h)

		#plt.plot((pp[0]+pp[1])/2,h,'o',color=col[c],ms=4,zorder=1000)

		pos[c] = (pp[0]+pp[1])/2,h
		#print "ciao"

	axarr[0].set_yticks(np.arange(0,1.2,0.2))
	axarr[0].set_ylim([0,max(hm)+0.05])

	axarr[0].yaxis.tick_right()
	axarr[0].tick_params(labelsize=14)
	axarr[0].set_ylabel('Distance',fontsize=18)
	axarr[0].yaxis.set_label_position("right")

	if nan==False:
		R = np.corrcoef(x)
	else:
		R =  np.array(pd.DataFrame(x.T).corr()) 

	ims = axarr[1].imshow(R,vmin=-1,vmax=1,cmap=cm.RdBu_r,interpolation='nearest')


	cbar_ax = fig.add_axes([0.95, 0.133, 0.03, 0.4])
	cbr = fig.colorbar(ims, cax=cbar_ax)

	cbr.set_label('correlation',fontsize=18)
	axarr[1].set_xlabel(xylabel,fontsize=18)
	axarr[1].set_ylabel(xylabel,fontsize=18)
	axarr[1].tick_params(labelsize=14)

	axarr[1].set_ylim([0,N])

	for l in L:
		if l == tuple(range(N)): continue

		axarr[1].add_patch(
			patches.Rectangle(
				(min(l)-0.5, min(l)-0.5),
				len(l),
				len(l),
				fill=False,
				lw=1.,
				color='green',	
				)
			)

	plt.setp(axarr[0].get_xticklabels(), visible=False)
	plt.subplots_adjust(hspace=.0)


	plt.savefig(file_w+'.pdf',bbox_inches='tight')

